from py4j.java_gateway import JavaObject
from typing import Generator


def iterate_java_object(iterable: JavaObject) -> Generator[JavaObject, None, None]:
    it = iterable.iterator()
    while it.hasNext():
        yield it.next()



import enum, dataclasses
from typing import List, Dict, Any, Generator, Optional


class TransformationType(enum.Enum):
    Alias = "alias"
    Project = "project"
    Filter = "filter"
    Join = "join"
    Source = "source"
    LogicalRDD = "logical_rdd"
    Relation = "relation"
    Generate = "generate"
    Aggregate = "aggregate"
    Window = "window"
    Sort = "sort"
    Union = "union"
    Limit = "limit"
    Repartition = "repartition"
    Deduplicate = "deduplicate"
    Except = "except"
    Intersect = "intersect"
    Sample = "sample"
    Expand = "expand"


Metadata = Dict[str, Any]


@dataclasses.dataclass
class TransformationColumn(object):
    name: str
    id: int
    type: str
    node_id: Optional[int]
    tree_string: str

    # List of the columns in previous nodes which form this column. This can be used
    # to navigate recursively until the data sources and discover its origins.
    links: List['TransformationColumn']


@dataclasses.dataclass
class TransformationNode:
    type: TransformationType
    metadata: Metadata  # TODO: make it more transformation-type specific
    children: List['TransformationNode']
    columns: Dict[int, TransformationColumn]

    @property
    def id(self) -> int:
        return id(self)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TransformationNode):
            return False

        return self.type == other.type and self.metadata == other.metadata and self.children == other.children

    def __hash__(self) -> int:
        return hash(id(self))

    def pprint(self, indent: int = 0) -> None:
        """Prints a human readable representation of the tree."""

        print("  " * indent + f"{self.type.name} {self.metadata}")
        for child in self.children:
            child.pprint(indent=indent + 1)


@dataclasses.dataclass
class JoinCondition:
    # SQL string representation of the condition.
    sql_str: str
    # Tree string representation of the condition.
    tree_string: str
    # The join condition is an equality of two columns with the same
    # name, or a conjunction of such equalities.
    # For example: 
    # - a = a -> True
    # - a > a -> False
    # - a = a AND b = b -> True
    # - a1 = a2 AND b1 = b2 -> False
    is_equi_join: bool
    # If is_equi_join is True, this dictionary contains a mapping from 
    # the names of the columns (used in the equi join condition), to the
    # column ids. Otherwise, this is an empty dictionary
    equi_join_columns: Dict[str, List[int]]



from py4j.java_gateway import JavaObject
from pyspark.sql import DataFrame
from typing import List, Dict, Any, Generator, Optional

from .dag_simplification import simplify
from .transformations_dag import TransformationNode
from .transformation_node_builders import build_dag_from_java_object
from .py4j_utils import iterate_java_object


def build_dag(df: DataFrame, simplify_dag: bool=True) -> TransformationNode:
    first_child = _get_last_transformation(df)
    raw_dag = build_dag_from_java_object(node=first_child)
    return simplify(raw_dag) if simplify_dag else raw_dag


def _get_last_transformation(df: DataFrame) -> JavaObject:
    # we add a dummy projection of all columns to Spark resolves the next
    # transformations in the plan. Otherwise, the columns will be "unresolved"
    df = df.select("*")
    root = df._jdf.queryExecution().logical()

    # skip the 'Project [*]' node
    return next(iterate_java_object(root.children()))



from .transformations_dag import Metadata, TransformationColumn, TransformationNode, TransformationType, JoinCondition

from typing import Dict, List, Set, TypeVar


class DagSimplifier(object):
    """
    The DAG Simplifiers detect certain pattern in a DAG, and map them to
    a simplified version of it.
    For example, four filters in a row could be simplified as one filter
    with four conditions.
    """

    def can_apply(self, dag: TransformationNode) -> bool:
        """Determines whether a pattern can be simplified by this simplifier or not"""
        raise NotImplementedError("Abstract method")
    
    def apply(self, original: TransformationNode) -> TransformationNode:
        """Returns a simplified version of the original DAG"""
        raise NotImplementedError("Abstract method")


class MergeJoinAndProject(DagSimplifier):
    """
    When the PySpark dataframe has a join, the resulting DAG has
    two nodes: one is a Join, and the other is a Project. With this simplification
    both nodes are merged.
    When joining by column name, the node has one column representing the
    column coming from the "left" df, and another for the "right" df. Our
    resulting merged join node contains only one column that has links
    to the two origin columns.
    """

    def __init__(self) -> None:
        self.already_merged_joins: Set[TransformationNode] = set()

    def can_apply(self, dag: TransformationNode) -> bool:
        if dag.type != TransformationType.Project:
            return False
        project_node = dag
        if len(project_node.children) != 1:
            return False
        join_node = project_node.children[0]
        if join_node.type != TransformationType.Join:
            return False
        if join_node.metadata["join_type"] == "Cross":
            # Cross Join doesn't add the extra Project, so removing it would be wrong
            return False
        if join_node in self.already_merged_joins:
            # Avoid collapsing all subsequent Projects
            return False
        condition = join_node.metadata["condition"]
        if not condition.is_equi_join:
            return False
        project_column_names = [c.name for c in project_node.columns.values()]
        join_column_names = [c.name for c in join_node.columns.values()]
        collapsed_col_names = [n for n in condition.equi_join_columns.keys()]
        return sorted(project_column_names + collapsed_col_names) == sorted(join_column_names)

    def apply(self, original: TransformationNode) -> TransformationNode:
        project_node = original
        join_node = original.children[0]
        new_node = TransformationNode(
            type=TransformationType.Join,
            metadata=self._merge_metadata(project_node.metadata, join_node.metadata),
            children=join_node.children,
            columns=self._merge_columns(project_node.columns, join_node.columns, join_node.metadata["condition"]),
        )
        for c in new_node.columns.values():
            c.node_id = new_node.id
        self.already_merged_joins.add(new_node)
        return new_node

    def _merge_metadata(self, proj_meta: Metadata, join_meta: Metadata) -> Metadata:
        return {
            "condition": join_meta["condition"],
            "join_type": join_meta["join_type"],
            "schema_string": proj_meta["schema_string"],
        }

    def _merge_columns(self, pcols: Dict[int, TransformationColumn], jcols: Dict[int, TransformationColumn], cond: JoinCondition) -> Dict[int, TransformationColumn]:
        for col in pcols.values():
            assert len(col.links) == 1
            link = col.links[0]

            links = []
            tree_string = ""
            if link.name in cond.equi_join_columns.keys():
                links = flatten([jcols[i].links for i in cond.equi_join_columns[link.name]])
                tree_string = cond.tree_string
            else:
                links = jcols[link.id].links
                tree_string = col.tree_string

            col.links = links
            col.tree_string = tree_string

        return pcols

class MergeDatasourceAndAlias(DagSimplifier):
    """
    When the PySpark dataframe represents a named table, it generates
    two nodes: one containing the table data itself, and another one
    containing an alias. This simplifier unifies both adding an
    attribute to the table node.
    When the datasource is a logical RDD, the alias is not added, so
    this simplifier considers that case and skips it.
    """

    def __init__(self) -> None:
        self.already_merged_relations: Set[TransformationNode] = set()

    def can_apply(self, df: TransformationNode) -> bool:
        if df.type != TransformationType.Alias:
            return False
        alias_node = df
        if alias_node in self.already_merged_relations:
            return False
        if len(alias_node.children) != 1:
            return False
        relation_node = alias_node.children[0]
        if relation_node.type != TransformationType.Relation:
            return False
        return True

    def apply(self, original: TransformationNode) -> TransformationNode:
        alias_node = original
        relation_node = original.children[0]
        new_node = TransformationNode(
            type=TransformationType.Relation,
            metadata=relation_node.metadata,
            children=relation_node.children,
            columns=self._remove_links(alias_node.columns),
        )
        for c in new_node.columns.values():
            c.node_id = new_node.id
        self.already_merged_relations.add(new_node)
        return new_node

    def _remove_links(self, alias_cols: Dict[int, TransformationColumn]) -> Dict[int, TransformationColumn]:
        for col in alias_cols.values():
            col.links = []
        return alias_cols


T = TypeVar("T")
def flatten(l: List[List[T]]) -> List[T]:
    return [item for sublist in l for item in sublist]


all_heuristics = [
    MergeJoinAndProject(),
    MergeDatasourceAndAlias(),
]


def simplify(raw_dag: TransformationNode) -> TransformationNode:
    # heuristics are applied with a "child-first" policy
    raw_dag.children = [simplify(child) for child in raw_dag.children]

    result_dag = raw_dag
    for heuristic in all_heuristics:
        if heuristic.can_apply(result_dag):
            result_dag = heuristic.apply(result_dag)
    return result_dag



from py4j.java_gateway import JavaObject
from typing import List, Dict, Any, Generator, Optional, Tuple

from .transformations_dag import JoinCondition, Metadata, TransformationColumn, TransformationNode, TransformationType
from .py4j_utils import iterate_java_object


class TransformationNodeBuilder(object):

    def build(self, java_node: JavaObject) -> TransformationNode:
        if self._expected_number_of_nodes() is not None:
            assert java_node.children().size() == self._expected_number_of_nodes(), java_node.children().size()

        metadata: Dict[str, str] = {}
        self._extract_common_metadata(java_node, metadata)
        self._extract_metadata(java_node, metadata)

        children = [build_dag_from_java_object(child) for child in iterate_java_object(java_node.children())]
        columns = self._extract_columns(java_node, children)

        node = TransformationNode(
            type=self._get_type(),
            metadata=metadata,
            children=children,
            columns=columns
        )
        for column in columns.values():
            column.node_id = node.id

        return node

    def _get_columns(self, java_node: JavaObject) -> JavaObject:
        return java_node.output()

    def _extract_columns(self, java_node: JavaObject, children: List[TransformationNode]) -> Dict[int, TransformationColumn]:
        columns = {}
        for java_column in iterate_java_object(self._get_columns(java_node)):
            column_id = java_column.exprId().id()
            parsed_column = self._extract_column(java_column, children)
            columns[column_id] = parsed_column
        return columns

    def _extract_column(self, java_column: JavaObject, children: List[TransformationNode]) -> TransformationColumn:
        if java_column.nodeName() not in ["Alias", "AttributeReference"]:
            raise NotImplementedError(f"Column type not supported: {java_column.nodeName()}")

        column_name = java_column.name()
        column_id = java_column.exprId().id()
        data_type = java_column.dataType().simpleString()
        reference_ids = [reference.exprId().id() for reference in iterate_java_object(java_column.references())]
        tree_string = java_column.treeString()

        # For each referenced column, search it in the children in order to add a link
        links = []
        for ref_id in reference_ids:
            for child in children:
                if ref_id in child.columns:
                    links.append(child.columns[ref_id])

        return TransformationColumn(
            name=column_name,
            id=column_id,
            type=data_type,
            node_id=None,
            links=links,
            tree_string=tree_string
        )

    def _extract_common_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["schema_string"] = node.schemaString()

    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        pass

    def _get_type(self) -> TransformationType:
        raise NotImplementedError("Abstract method")

    def _expected_number_of_nodes(self) -> Optional[int]:
        raise NotImplementedError("Abstract method")


class ProjectNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        # TODO: Something besides the columns?
        pass

    def _get_columns(self, node: JavaObject) -> JavaObject:
        return node.projectList()

    def _get_type(self) -> TransformationType:
        return TransformationType.Project

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1


class FilterNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["condition"] = node.condition().sql()

    def _get_type(self) -> TransformationType:
        return TransformationType.Filter

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1


class LogicalRDDNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        # TODO: collect metadata about RDD columns
        metadata["type"] = "Logical RDD"

    def _get_type(self) -> TransformationType:
        return TransformationType.LogicalRDD

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 0


class LogicalRelationNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["type"] = "relation"
        metadata["file_format"] = node.relation().fileFormat().toString()

        catalog_table_option = node.catalogTable()
        if not catalog_table_option.isEmpty():
            catalog_table = catalog_table_option.get()
            metadata["database"] = catalog_table.database()
            metadata["table"] = catalog_table.identifier().table()
            metadata["partition_columns"] = [name for name in iterate_java_object(catalog_table.partitionColumnNames())]
            metadata["storage"] = catalog_table.storage().toString()

    def _get_type(self) -> TransformationType:
        return TransformationType.Relation

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 0


class JoinNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["join_type"] = node.joinType().toString()
        metadata["condition"] = self._extract_condition(node)

    def _get_type(self) -> TransformationType:
        return TransformationType.Join

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 2

    def _extract_condition(self, node: JavaObject) -> JoinCondition:
        if node.condition().isEmpty():
            return JoinCondition("", "", False, {})
        cond = node.condition().get()
        is_equi_join, equi_join_columns = self._is_equi_join(cond)
        reference_ids = [ref.exprId().id() for ref in iterate_java_object(cond.references())]
        return JoinCondition(cond.sql(), cond.treeString(), is_equi_join, equi_join_columns)

    def _get_column_name(self, java_object: JavaObject) -> Optional[str]:
        if java_object.nodeName() == "AttributeReference":
            return str(java_object.name())
        elif java_object.nodeName() == "Cast":
            return self._get_column_name(java_object.child())
        else:
            return None

    def _is_equi_join(self, cond: JavaObject) -> Tuple[bool, Dict[str, List[int]]]:
        node_name = cond.nodeName()
        if node_name == "EqualTo":
            cols = [col for col in iterate_java_object(cond.children())]
            reference_ids = [ref.exprId().id() for ref in iterate_java_object(cond.references())]
            if len(cols) != 2:
                return False, {}
            name1 = self._get_column_name(cols[0])
            name2 = self._get_column_name(cols[1])
            if name1 is not None and name1 == name2:
                return True, {name1: reference_ids}
            else:
                return False, {}

        if node_name == "And":
            equi_join_columns = {}
            for child in iterate_java_object(cond.children()):
                child_is_equi, child_cols = self._is_equi_join(child)
                if child_is_equi:
                    equi_join_columns.update(child_cols)
                else:
                    return False, {}
            return True, equi_join_columns

        return False, {}


class GenerateNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["generator"] = node.generator().sql()

    def _get_type(self) -> TransformationType:
        return TransformationType.Generate

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1


class AggregateNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        aggregate_expressions = []
        for aggregate_expression in iterate_java_object(node.aggregateExpressions()):
            aggregate_expressions.append(aggregate_expression.sql())

        grouping_expressions = []
        for grouping_expression in iterate_java_object(node.groupingExpressions()):
            grouping_expressions.append(grouping_expression.sql())

        metadata['aggregate_expressions'] = aggregate_expressions
        metadata['grouping_expressions'] = grouping_expressions

    def _get_type(self) -> TransformationType:
        return TransformationType.Aggregate

    def _get_columns(self, java_node: JavaObject) -> JavaObject:
        return java_node.aggregateExpressions()

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1


class WindowNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        # TODO: parse metadata
        pass

    def _get_type(self) -> TransformationType:
        return TransformationType.Window

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1


class SortNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["order"] = [c.sql() for c in iterate_java_object(node.order())]

    def _get_type(self) -> TransformationType:
        return TransformationType.Sort

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1


class UnionNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        # Nothing to extract here
        pass

    def _get_type(self) -> TransformationType:
        return TransformationType.Union

    def _expected_number_of_nodes(self) -> Optional[int]:
        # Might be any number greater than 1
        return None


class LimitNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["limit_expr"] = node.limitExpr().toString()

    def _get_type(self) -> TransformationType:
        return TransformationType.Limit

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1


class AliasNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["alias"] = node.identifier().toString()

    def _get_type(self) -> TransformationType:
        return TransformationType.Alias

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1


class RepartitionNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["num_partitions"] = node.numPartitions()
        metadata["expressions"] = [exp.sql() for exp in iterate_java_object(node.expressions())]

    def _get_type(self) -> TransformationType:
        return TransformationType.Repartition

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1


class DeduplicateNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["deduplicate_columns"] = [col.sql() for col in iterate_java_object(node.references())]

    def _get_type(self) -> TransformationType:
        return TransformationType.Deduplicate

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1


class ExceptAllNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["preserves_duplicates"] = True

    def _get_type(self) -> TransformationType:
        return TransformationType.Except

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 2


class ExceptNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["preserves_duplicates"] = False

    def _get_type(self) -> TransformationType:
        return TransformationType.Except

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 2


class IntersectNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["preserves_duplicates"] = False

    def _get_type(self) -> TransformationType:
        return TransformationType.Intersect

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 2


class IntersectAllNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["preserves_duplicates"] = True

    def _get_type(self) -> TransformationType:
        return TransformationType.Intersect

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 2


class SampleNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["seed"] = node.seed()
        metadata["fraction"] = node.fraction()


    def _get_type(self) -> TransformationType:
        return TransformationType.Sample

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1


class ExpandNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["projections"] = [self._proj(proj) for proj in iterate_java_object(node.projections())]

    def _proj(self, proj: List[JavaObject]) -> List[str]:
        return [col.sql() for col in iterate_java_object(proj)]

    def _get_type(self) -> TransformationType:
        return TransformationType.Expand

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1


def build_dag_from_java_object(node: JavaObject) -> TransformationNode:
    # objects that can parse each transformation sub-tree
    transformation_builders: Dict[str, 'TransformationNodeBuilder'] = {
        "SubqueryAlias": AliasNodeBuilder(),
        "Project": ProjectNodeBuilder(),
        "Filter": FilterNodeBuilder(),
        "LogicalRDD": LogicalRDDNodeBuilder(),
        "LogicalRelation": LogicalRelationNodeBuilder(),
        "Join": JoinNodeBuilder(),
        "Generate": GenerateNodeBuilder(),
        "Aggregate": AggregateNodeBuilder(),
        "Window": WindowNodeBuilder(),
        "Sort": SortNodeBuilder(),
        "Union": UnionNodeBuilder(),
        "GlobalLimit": LimitNodeBuilder(),
        "LocalLimit": LimitNodeBuilder(),
        "Repartition": RepartitionNodeBuilder(),
        "Deduplicate": DeduplicateNodeBuilder(),
        "Except": ExceptNodeBuilder(),
        "Except All": ExceptAllNodeBuilder(),
        "Intersect": IntersectNodeBuilder(),
        "Intersect All": IntersectAllNodeBuilder(),
        "Sample": SampleNodeBuilder(),
        "RepartitionByExpression": RepartitionNodeBuilder(),
        "Expand": ExpandNodeBuilder(),
    }

    builder = transformation_builders.get(node.nodeName())
    if not builder:
        raise NotImplementedError(f"Transformation not supported: '{node.nodeName()}'")

    return builder.build(node)



import os
import shutil
import json
import dataclasses

from .default_settings import DefaultSettings as DefaultSettings  # explicit re-export for mypy
from .plan_extractor import dag
from .plan_extractor.dag_builder import build_dag
from .plan_extractor.transformations_dag import JoinCondition, TransformationColumn, TransformationNode, TransformationType

from pyspark.sql import DataFrame
from typing import Dict, Any, List, Tuple


# string template to build the JS file containing the graph definition as nodes
# and links objects required by the framework that will draw them
MODEL_FILE_TEMPLATE = """
const model_defaultSettings = {settings};

const model_initialNodes = {nodes};

const model_initialEdges = {links};
"""

class Encoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if type(o) is JoinCondition:
            return dataclasses.asdict(o)
        return json.JSONEncoder.default(self, o)    


def dump_dataframe(df: DataFrame, output_dir: str, overwrite: bool, default_settings: DefaultSettings, simplify_dag: bool=True) -> None:
    """Create a visual representation of the given `dag` in HTML. The HTML
    files will be saved in the `output_dir` directory. If `overwrite` is
    True, the output directory will be deleted if it already exists.
    If `simplify_dag` is True, a series of heuristics will be applied to
    simplify the resulting DAG."""

    tree = build_dag(df=df, simplify_dag=simplify_dag)
    nodes, links = get_nodes_and_links(tree)

    model_file = MODEL_FILE_TEMPLATE.format(
        nodes=json.dumps(nodes, indent=4, cls=Encoder),
        links=json.dumps(links, indent=4),
        settings=json.dumps(dataclasses.asdict(default_settings), indent=4),
    )

    if overwrite and os.path.exists(output_dir):
        shutil.rmtree(output_dir)

    # copy the ui directory to the output directory
    here = os.path.dirname(__file__)
    shutil.copytree(os.path.join(here, "ui"), output_dir)

    # create the output file with the model data
    with open(os.path.join(output_dir, "model.js"), "w") as fp:
        fp.write(model_file)


def _get_column_id(column: TransformationColumn) -> str:
    return f"{str(column.node_id)}->{str(column.id)}"


def _column_as_dict(column: TransformationColumn) -> Dict[str, object]:
    return {
        "id": _get_column_id(column),
        "type": "column",
        "parentNode": str(column.node_id),
        "expandParent": True,
        "data": {
            "id": column.id,
            "name": column.name,
            "type": column.type,
            "tree_string": column.tree_string,
            "linked_columns": [_get_column_id(link) for link in column.links]
        },
    } 


def _transformation_as_dict(node: TransformationNode) -> Dict[str, object]:
    # map the Node.type to the type required by the HTML DAG renderer
    node_type_map = {
        TransformationType.Alias: 'Alias',
        TransformationType.Project: 'Project',
        TransformationType.Filter: 'Filter',
        TransformationType.LogicalRDD: 'DataSource',
        TransformationType.Relation: 'DataSource',
        TransformationType.Generate: 'Transform',
        TransformationType.Aggregate: 'Group',
        TransformationType.Join: 'Join',
        TransformationType.Sort: 'Sort',
        TransformationType.Window: 'Window',
        TransformationType.Union: 'Union',
        TransformationType.Limit: 'Limit',
        TransformationType.Repartition: 'Repartition',
        TransformationType.Deduplicate: 'Deduplicate',
        TransformationType.Except: 'Except',
        TransformationType.Intersect: 'Intersect',
        TransformationType.Sample: 'Sample',
        TransformationType.Expand: 'Expand',
    }

    return {
        # each node must have a unique ID
        "id": str(node.id),
        # the type of the node for react-flow to use the correct component
        "type": "transformation",
        # data passed to the node upon creation
        "data": {
            # store the transformation type
            "type": node_type_map[node.type],
            # label to display in the node
            "label": node_type_map[node.type],
            # Metadata depending on node type
            "metadata": node.metadata,
            "columns": [_get_column_id(column) for column in node.columns.values()],
        },
    }


def _transformation_link_as_dict(data: dag.DFSNodeData, node: TransformationNode) -> Dict[str, object]:
    return {
        "id": f"{data.parent_id}-{node.id}",
        "source": str(data.parent_id),
        "target": str(node.id),
    }


def _column_link_as_dict(source_column: TransformationColumn, target_column: TransformationColumn) -> Dict[str, Any]:
    source = f"{str(source_column.node_id)}->{str(source_column.id)}"
    target = f"{str(target_column.node_id)}->{str(target_column.id)}"
    return {
        "id": f"({source})-({target})",
        "source": source,
        "target": target,
        "animated": True,
    }


def get_nodes_and_links(tree: TransformationNode) -> Tuple[List[Any], List[Any]]:
    """Convert the tree into a list of nodes and links for the HTML DAG."""

    transformation_nodes, transformation_links, column_nodes, column_links = [], [], [], []
    for data, node in dag.dfs(tree):
        transformation_nodes.append(_transformation_as_dict(node))
        # if the parent is None, it means we are at the root of the tree
        if data.parent_id is not None:
            transformation_links.append(_transformation_link_as_dict(data, node))

        for column in node.columns.values():
            column_nodes.append(_column_as_dict(column))
            for link in column.links:
                column_links.append(_column_link_as_dict(column, link))

    return transformation_nodes + column_nodes, transformation_links + column_links
